// Canvas WorldBuilder
// Eric Turner
// 2011

var BUILDER = (function(window, undefined) {

    var me = {},
		canvas,         // The canvas element.
		ctx,            // The actual 2D context we can draw to.
		mouse_text,     // Textual coordinates of the relative mouse x,y.
		grid_text,      // Textual coordinates of row and column.
		images = [],    // Available tiles.
        tiles = [],     // A collection of tile elements within the map.
        selected_tile = undefined,      // the currently selected tile to paint with

		COLS = 10,
		ROWS = 10,

		TILEBAR_WIDTH = 7,

		IMAGE_MOUSEOVER_CLASS = 'mouse_over',
		IMAGE_MOUSEOUT_CLASS = 'mouse_out',

		COLOR_GRAY = '#CCC',
		COLOR_BLACK = '#000',
		COLOR_WHITE = '#FFF',
		COLOR_RED = '#FF0000',
		COLOR_GREEN = '#00FF00',
		COLOR_BLUE = '#0000FF',

		DEFAULT_HEIGHT = 480,
		DEFAULT_WIDTH = 480,
        DEFAULT_CONTEXT = '2d',
        
        
        _end = null;   //----------------------------------------------------

	/*
    TODO: In a finished version there could be a lot of Tile objects created,
    so it could eventually be optimized to define functions through 
    Tile.prototype. */
	var Tile = function(x, y, img) {
		if (this instanceof Tile) {
			this.x = x;
			this.y = y;
			
			/* We'll tolerate an Image or just a url string. */
			if (img instanceof Image) {
				this.img = img; 
			} else if (typeof img == 'string') {
				this.img = new Image();
				this.img.src = img;
			}

			this.draw = function() {
				ctx.drawImage(this.img, this.x, this.y);
			}

		} else {
			return new Tile(x, y, img);
		}
		return this;
	}

	var TileSet = (function() {
		var me = {},
			set = [];
		me.add = function(tile) {
		    set.push(tile);
        }


		return me;
	})();

	var Cursor = (function() {
		var me = {},
			x,
			y;

        me.set_pos = function(x, y) {
			this.x = x * (DEFAULT_WIDTH / COLS) + .5;
			this.y = y * (DEFAULT_HEIGHT / ROWS) + .5;
		};

		me.get_x = function() {
			return this.x;
		};

		me.get_y = function() {
			return this.y;
		};

		me.get_width = function() {
			return DEFAULT_WIDTH / COLS;
		};

		me.get_height = function() {
			return DEFAULT_HEIGHT / ROWS;
		};

		return me;
	})();

	function preload(assets, func) {
		var i = 0,
			to_load = 0,
			loaded = 0;
		for(i = 0; i < assets.length; ++i) {
			++to_load;
			var img = new Image();
			img.src = assets[i];
			img.onload = function() {
				++loaded;
			};

			images.push(img);
		}

		/*  This will run every 100ms until all assets are loaded, then
			invoke the callback given by func
		*/
		(function relax(){
			if (to_load > loaded) {
				setTimeout(relax, 100);
			} else {
				func();
			}
		})();
	}

    // Build the UI
	function run() {
		ctx = canvas.getContext(DEFAULT_CONTEXT);
		ctx.fillStyle = COLOR_GRAY;
		ctx.fillRect(0, 0, DEFAULT_WIDTH, DEFAULT_HEIGHT);
	
		canvas.onmouseover = mouse_over;
		canvas.onmouseout = mouse_out;
        canvas.onmousedown = mouse_down;
		canvas.onmousemove = mouse_move;

		draw_grid();
		build_tile_bar();
	}

	// Draw the actual map
    function draw_grid() {
		var i = 0;
		

		ctx.fillStyle = COLOR_GRAY;
		ctx.fillRect(0, 0, DEFAULT_WIDTH, DEFAULT_HEIGHT);
       
        // Draw the tiles
        for (i = 0; i < tiles.length; ++i) {
            tiles[i].draw();
        }
		
		for (i = DEFAULT_WIDTH / COLS; i <= DEFAULT_WIDTH; i += DEFAULT_WIDTH / COLS) {
			ctx.moveTo(i + .5, 0);
			ctx.lineTo(i + .5, DEFAULT_HEIGHT);
		}

		for (i = DEFAULT_HEIGHT / ROWS; i <= DEFAULT_HEIGHT; i += DEFAULT_HEIGHT / ROWS) {
			ctx.moveTo(0, i + .5);
			ctx.lineTo(DEFAULT_WIDTH, i + .5);
		}

		ctx.strokeStyle = '#000';
		ctx.stroke();
	}

    // Update the text lable with the mouse coordinates
	function update_mouse(x, y) {
		if (!mouse_text) {
			mouse_text = document.createElement('p');
			document.body.appendChild(mouse_text);
		}

		while (mouse_text.firstChild !== null)
			mouse_text.removeChild(mouse_text.firstChild);
		mouse_text.appendChild(document.createTextNode('(' + x + ', ' + y + ')'));
	}
	
    // Update the text lable with the grid coordinates
	function update_grid(x, y) {
		if (!grid_text) {
			grid_text = document.createElement('p');
			document.body.appendChild(grid_text);
		}

		while (grid_text.firstChild !== null)
			grid_text.removeChild(grid_text.firstChild);
		
        x = x || '?';
        y = y || '?';

        
        grid_text.appendChild(document.createTextNode('(' + x + ', ' + y + ')'));

		draw_grid();
	}

    // Draw the cursor to the grid
	function update_cursor(x, y) {
		Cursor.set_pos(x, y);
		
        ctx.strokeStyle = '#FF0000';

        if (selected_tile !== undefined) {
            ctx.drawImage(selected_tile, Cursor.get_x(), Cursor.get_y());
        }
		
        ctx.strokeRect(Cursor.get_x() + 1, Cursor.get_y() + 1
			, Cursor.get_width() - 2, Cursor.get_height() - 2);
	}

    // Construct the set of images that we'll use to make maps
	function build_tile_bar() {
		document.body.appendChild(document.createElement('br'));
		for (var i = 0; i < images.length; ++i) {
			
			images[i].setAttribute('id', 'image' + i);
			images[i].className = IMAGE_MOUSEOUT_CLASS;

			if (i % TILEBAR_WIDTH == 0) {
				document.body.appendChild(document.createElement('br'));
			}

			document.body.appendChild(images[i]);

            // Bind the mouse events to this tile
			images[i].onmouseover = mouse_over;
            images[i].onmouseout = mouse_out;
            images[i].onmousedown = mouse_down;

		}
	}

    /* Mouse Events */

    function mouse_down(e) {
        var e = e || window.event,
            target = e.target || e.srcElement,
            grid_x = 0,
            grid_y = 0;
        
        if (target.tagName == 'IMG') {
            //console.log(target.getAttribute('src'));
            
            if (selected_tile !== undefined) {
                selected_tile.className = IMAGE_MOUSEOUT_CLASS; 
            }

            selected_tile = target;
            target.className = IMAGE_MOUSEOVER_CLASS;
        } else if (target.tagName == 'CANVAS' && selected_tile != undefined) {
            tiles.push(new Tile(Cursor.get_x(), Cursor.get_y(), selected_tile));
            //console.log('Tile placed on: (' + Cursor.get_x() + ', ' + Cursor.get_y() + ' )');
        }
    }

	function mouse_over(e) {
		var e = e || window.event,
            target = e.target || e.srcElement;

        if (target.tagName == 'IMG') {
            target.className = IMAGE_MOUSEOVER_CLASS;
        }
	}

	function mouse_out(e) {
		var e = e || window.event,
            target = e.target || e.srcElement;

        if (target.tagName == 'IMG' && target !== selected_tile) {
            target.className = IMAGE_MOUSEOUT_CLASS;
        } else if (target.tagName == 'CANVAS') {
            update_grid();
        }
    }

	function mouse_move(e) {
		if (!e) var e = window.event;
		
		var i,
			k,
			pos_x = 0,
			pos_y = 0,
			grid_x = 0,
			grid_y = 0;

		if (e.pageX || e.pageY) {
			pos_x = e.pageX;
			pos_y = e.pageY;
		} else if (e.clientX || e.clientY) {
			pos_x = e.clientX + document.body.scrollLeft
				+ document.documentElement.scrollLeft;
			pos_y = e.clientY + document.body.scrollTop
				+ document.documentElement.scrollTop;
		}

		/* Fixes some weird bugs that can happen with you move near the borders... 
         * Needs more testing */
		if (pos_x < 0)
			pos_x = 0;
		if (pos_y < 0)
			pos_y = 0;
		if (pos_x > canvas.offsetLeft + DEFAULT_WIDTH)
			pos_x = canvas.offsetLeft + DEFAULT_WIDTH;
		if (pos_y > canvas.offsetTop + DEFAULT_HEIGHT)
			pos_y = canvas.offsetTop + DEFAULT_HEIGHT;


		/* Just loop through the grid to find out where the mouse is. */
		//TODO: this only uses offsetTop / offsetLeft and isn't yet scroll safe
		//console.log(canvas.offsetTop);
		for (i = canvas.scrollLeft, k = 0; i <= (DEFAULT_WIDTH + canvas.offsetLeft); 
			i += DEFAULT_WIDTH / COLS, ++k) {
			if (pos_x >= i - DEFAULT_WIDTH / COLS && pos_x <= i) {
				grid_x = k - 1;
			}
		}
		for (i = canvas.offsetTop, k = 0; i <= (DEFAULT_HEIGHT + canvas.offsetTop); 
			i += DEFAULT_HEIGHT / ROWS, ++k) {
			if (pos_y >= i - DEFAULT_HEIGHT / ROWS && pos_y <= i) {
				grid_y = k - 1;
			}
		}
		
		//console.log('(' + pos_x + ', ' + pos_y + ')');

		update_mouse(pos_x, pos_y);
		update_grid(grid_x, grid_y);
		update_cursor(grid_x, grid_y);
	}

	/*  
		We'll either build the actual map editor, or notify the user that
		they need a browser that can do canvas.
	*/
	me.build = function() {
		canvas = document.createElement('canvas');

		if (canvas) {
			canvas.setAttribute('width', DEFAULT_WIDTH);
			canvas.setAttribute('height', DEFAULT_HEIGHT);
			document.body.appendChild(canvas);
			
			preload(['./assets/grass1.png', './assets/grass2.png'
				, './assets/grass3.png', './assets/grass4.png'
				, './assets/ground1.png', './assets/ground2.png'
				, './assets/ground3.png', './assets/ground4.png'
				, './assets/ground5.png', './assets/ground6.png'
				, './assets/ground7.png', './assets/ground8.png'],  run);


		} else {
			canvas = document.createElement('div');
			canvas.setAttribute('id', 'no_canvas');
			document.body.appendChild(canvas);
		}
	};

	return me;

})(window);

window.onload = BUILDER.build;
